/********************************************************
 * ex-test -- Test a exabyte tape drive.		*
 *							*
 * Usage:						*
 *	ex-test [-c<count>] [-b<block>] [-r] device	*
 *							*
 *		-c<count> -- Number of record to try	*
 *		-b<blocking> -- Block size to use	*
 *		-r -- Print raw log information		*
 *		-v -- Verbose				*
 *							*
 * $Header: $						*
 ********************************************************/
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <time.h>

#include <fcntl.h>
#include <sys/mtio.h>

#include <scsi/scsi.h>
#include <scsi_ioctl.h>

const int EX_LOG_SENSE = 0x4D;	// Read Log Sense 

const int READ_LOG = 0x43;	// The log page for reading
const int WRITE_LOG = 0x42;	// The write log page

static bool raw = false;	/* Output raw data */
static int fd;			/* File descriptor of the tape */
static bool verbose = false;	/* Verbose output */

struct log_info {
    unsigned char page_code;		/* Page code from the log */
    unsigned char reserved;		/* Reserved byte */
    unsigned char size_code[2];		/* Size of the data */
};
union scsi_log_struct {
    struct log_info info;	/* Information for a log */
    unsigned char raw[500];	/* Raw SCSI data */
};
/* The log data from the tape drive */
static union scsi_log_struct log;

/*
 * Information from the exabyte error log
 */
class error_information {
    public:
	int total_retries;	/* Total retries corrected (param code=2) */
	int total_corrections;	/* Total errors corrected (3) */
	int total_errors;	/* Total Errors processed (4) */
	long long int total_bytes;/* Total bytes processed (5) */
	int total_hard;		/* Total hard errors (pc=6) */
    public:
	error_information(void):
	    total_retries(0),
	    total_corrections(0),
	    total_errors(0),
	    total_bytes(0LL),
	    total_hard(0)
	{}
	// Destructor defaults
	// Copy constructor defaults
	// Assignment operator defaults
    public:
	void print(
	    const char what[]	// The name of the counter
	);
};

struct tape_info {
    long long int read_size;	// Size of the reads
    long long int write_size;	// Size of the writes
};

static FILE* tty_file = NULL;	// File for verbose I/O

/********************************************************
 * subtract two error_information structures		*
 ********************************************************/
static error_information operator - (
    const error_information& e1,
    const error_information& e2
) {
    error_information result;	// Result of the subraction

    result.total_retries = e1.total_retries - e2.total_retries;
    result.total_corrections = e1.total_corrections - e2.total_corrections;
    result.total_errors = e1.total_errors - e2.total_errors;
    result.total_bytes = e1.total_bytes - e2.total_bytes;
    result.total_hard = e1.total_hard - e2.total_hard;

    return (result);
}

/********************************************************
 * raw_page -- Print the page in the raw		*
 ********************************************************/
static void raw_page(void)
{
    int i;	/* Index for our dump */
    printf("----------------------------------------------------\n");

    for (i = 0; i < log.info.size_code[1]+4; ++i) {
	printf("%2d: 0x%02x\n", i, log.raw[i]);
    }
    printf("\n\n\n");
}

/* Turn variable into a signed character */
static inline signed char SC(
    const int what
) {
    return(static_cast<signed char>(what));
}

/* Turn variable into a unsigned long long */
#define ULL(value, index) \
     (static_cast<unsigned long long>((value)[index]))

/* Turn variable into a unsigned long */
#define UL(value, index) \
     (static_cast<unsigned long>((value)[index]))

/* Turn variable into a unsigned int */
#define U(value, index) \
     ((unsigned int)((value)[index]))

/* Turn variable into a unsigned int */
#define L(value, index) \
     ((unsigned int)(SC((value)[index])))

/* Turn two bytes into a a short int */
#define TWO_BYTES(value) (signed short int) (\
    ((U(value,0) << 8) | U(value, 1))	    )

/* Turn three bytes into a a short int */
#define THREE_BYTES(value) \
    ((L( value, 0) << 16L) | \
     (UL(value, 1) << 8L)  | \
      UL(value, 2))

/* Turn four bytes into a a short int */
#define FOUR_BYTES(value) \
   ((UL(value, 0) << 24L) | \
    (UL(value, 1) << 16L) | \
    (UL(value, 2) << 8L)  | \
     UL(value, 3))

#define FIVE_BYTES(value) \
   ((ULL(value, 0) << 32L) | \
    (ULL(value, 1) << 24L) | \
    (ULL(value, 2) << 16L) | \
    (ULL(value, 3) << 8L)  | \
     ULL(value, 4))

/********************************************************
 * get_page -- Get a page of the LOG_SENSE data		*
 *							*
 * Data is returned in the global structure log		*
 ********************************************************/
static void get_page(
    const int page	/* Page index */
)
{
    static struct log_cmd {
	unsigned int inlen;		 /* Length of data going in */
	unsigned int outlen;		 /* Length of data going out */
	unsigned char cmd[10];		 /* The command to execute */
	unsigned char data[sizeof(log)]; /* The data buffer */
    } scsi_cmd;

    scsi_cmd.cmd[0] = EX_LOG_SENSE;
    scsi_cmd.cmd[1] = 0;
    scsi_cmd.cmd[2] = page;

    scsi_cmd.cmd[3] = 0;
    scsi_cmd.cmd[4] = 0;
    scsi_cmd.cmd[5] = 0;
    scsi_cmd.cmd[6] = 0;

    scsi_cmd.cmd[7] = sizeof(log)/256;
    scsi_cmd.cmd[8] = sizeof(log)%256;
    scsi_cmd.cmd[9] = 0;

    scsi_cmd.inlen = 0;
    scsi_cmd.outlen = sizeof(log);

    if (ioctl (fd, SCSI_IOCTL_SEND_COMMAND, &scsi_cmd) < 0) { 
	int error = errno;
	perror("SCSI command failed: ");
	printf("Error code: %d\n", error);
	exit (1);
    }

    /* Copy the the data to the log variable */
    /* (I know it's a hack) */
    memcpy(&log, scsi_cmd.cmd, sizeof(log));
}

/********************************************************
 * report_errors -- print a error counter report	*
 *							*
 * Page format						*
 * 	0,1 -- page code				*
 * 	2 -- Flags					*
 *	3 -- page length (1c)				*
 *							*
 *	4,5 -- 0,1 param code (3)			*
 *	6 -- 2 flags					*
 *	7 -- 3 param length (2)				*
 *	8,9 -- 4,5 value				*
 *							*
 *	10,11 -- 0,1 param code (4)			*
 *	12 -- 2 flags					*
 *	13 -- 3 param length (4)			*
 *	14,15,16,17 -- 4,5,6,7 value			*
 *							*
 *	18,19 -- 0,1 param code (5)			*
 *	20 -- 2 flags					*
 *	21 -- 3 param length (4)			*
 *	22,23,24,25 -- 4,5,6,7 value			*
 *							*
 *	26,27 -- 0,1 param code (6)			*
 *	28 -- 2 flags					*
 *	29 -- 3 param length (2)			*
 *	30,31 -- 4,5 value				*
 ********************************************************/
static struct error_information report_error(
    const int page	/* The page number for the report */
)
{
    struct error_information result;

    get_page(page);
    if (raw) raw_page();

    result.total_retries     = THREE_BYTES(&log.raw[8]);
    result.total_corrections = THREE_BYTES(&log.raw[15]);
    result.total_errors      = THREE_BYTES(&log.raw[22]);
    result.total_bytes       = FIVE_BYTES(&log.raw[29]);
    result.total_hard        = TWO_BYTES(&log.raw[38]);

    if ((result.total_bytes &       0x8000000000LL) != 0LL) {
	result.total_bytes |= 0xFFFFFF8000000000LL;
    }
    return (result);
}
/********************************************************
 * print_errors -- print the error information		*
 ********************************************************/
void error_information::print(
    const char what[]	// The name of the counter
) {
    printf("%s errors: Bytes %5lld Retries %3d Corrections %3d Soft %3d Hard %3d\n",
    	what, 
	total_bytes,
	total_retries, 
	total_corrections,
	total_errors, 
	total_hard);
}

/********************************************************
 * rewind -- Rewind the tape.				*
 ********************************************************/
static void rewind_tape(
    const int tape_fd	// Fd of the tape to rewind
) {
    // How to tell the drive to rewind
    static struct mtop rew = {
	MTREW,
	1
    };

    if (ioctl(tape_fd, MTIOCTOP, &rew) != 0) {
	fprintf(stderr, "Unable to rewind tape drive: %d\n", errno);
	exit (8);
    }
}

/********************************************************
 * test_tape -- Test a tape drive			*
 ********************************************************/
static tape_info test_tape(
    const unsigned long int count,	// Number of blocks to try
    const unsigned int block_size	// Blocking factor
) {
    tape_info result;

    char *buffer = new char[block_size];
    if (buffer == NULL) {
	fprintf(stderr, "Error: Could not allocate buffer\n");
	exit (8);
    }

    unsigned long long int cur_block;	// Current block number

    rewind_tape(fd);
    for (cur_block = 0; cur_block < count; ++cur_block) {
	if (verbose) {
	    fprintf(tty_file, "%Ld\r", cur_block);
	    fflush(tty_file);
	}

	char *cur_byte;
	for (cur_byte = buffer;
	    cur_byte <= (buffer + block_size);
	    ++cur_byte) {
	    *cur_byte = lrand48();
	}

	*((unsigned long int *)buffer) = cur_block;

	if (write(fd, buffer, block_size) != static_cast<int>(block_size)) {
	    if (verbose) fprintf(tty_file, "\n");
	    printf("Write: Block %Lu Error %d\n", cur_block, errno);
	    break;
	}
    }
    if (verbose) fprintf(tty_file, "\n");
    rewind_tape(fd);
    unsigned long long int write_count = cur_block;

    for (cur_block = 0; cur_block < write_count; ++cur_block) {
	if (verbose) {
	    fprintf(tty_file, "%Ld\r", cur_block);
	    fflush(stdout);
	}

	if (read(fd, buffer, block_size) != static_cast<int>(block_size)) {
	    if (verbose) printf("\n");
	    printf("ERROR: Unable to read block %Ld Error %d\n", cur_block, errno);
	    break;
	}
    }
    if (verbose) fprintf(tty_file, "\n");
    delete[] buffer;
    result.read_size = cur_block * block_size;
    result.write_size = write_count * block_size;
    return (result);
}

/********************************************************
 * usage -- tell the user what to do			*
 ********************************************************/
static void usage(void)
{
    printf("ex-test [-c<count>] [-b<block>] [-r] device\n");
    printf("Options:\n");
    printf("	-c<count> -- Number of record to try\n");
    printf("	-b<blocking> -- Block size to use\n");
    printf("	-r -- Print raw log information\n");
    printf("	-v -- Verbose output\n");
    exit (8);
}


int main (int argc, char *argv[])
{ 
    // Blocking factor
    unsigned long int blocking = 10 * 1024;

    // Number of blocks to try test
    unsigned long int count = 0x7fffffffL;

    while ((argc > 1) && (argv[1][0] == '-')) {
	switch (argv[1][1]) {
	    case 'b':
		blocking = atol(&argv[1][2]);
		break;
	    case 'c':
		count = atol(&argv[1][2]);
		break;
	    case 'r':
		raw = true;
		break;
	    case 'v':
		verbose = true;
		break;
	    default:
		printf("Bad option %s\n", argv[1]);
		usage();
	}
	--argc;
	++argv;
    }
    if (argc != 2) {
        usage();
    }

    if (verbose) {
	tty_file = fopen("/dev/tty", "w");
	if (tty_file == NULL) {
	    fprintf(stderr, "Could not open /dev/tty\n");
	    exit (8);
	}
    }

    if ((fd = open(argv[1], O_RDWR | O_NONBLOCK)) < 0) { 
	if ((fd = open(argv[1], O_RDWR | O_NONBLOCK)) < 0) { 
	    printf ("Failed to open device %s\n", argv[1]);
	    exit (1);
	}
    }

    // Time we start the process
    time_t start_time = time(0);

    // Information before the test
    error_information b4_read = report_error(READ_LOG);
    error_information b4_write = report_error(WRITE_LOG);

    // The size of the tape
    struct tape_info tape_size = test_tape(count, blocking);

    // Information after the test
    error_information after_read = report_error(READ_LOG);
    error_information after_write = report_error(WRITE_LOG);

    // The time things finish up
    time_t end_time = time(0);

    if (verbose) {
	after_read.print(            "Read  After");
	b4_read.print(               "Read  Before");
	(after_read - b4_read).print("Read  Result");

	printf("\n");
	after_write.print(             "Write After");
	b4_write.print(                "Write Before");
	(after_write - b4_write).print("Write Result");
    } else {
	(after_read - b4_read).print("Read  Result");
	(after_write - b4_write).print("Write Result");
    }
    double mb_read = static_cast<double>(tape_size.read_size)
			/ (1024.0 * 1024.0);
    printf("Read %5.1fMB\n", mb_read);
    if (mb_read > 1000.0) 
	printf("Read %5.1fGB\n", mb_read / 1024.0);

    printf("Tape speed %5.1f KB/Second\n",
	    (static_cast<double>(tape_size.read_size) /
	         (1024.0) ) /
	    		static_cast<double>((end_time - start_time)/2));

    return(0);
}
